@@ -1,26 +1,27 @@
 # Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
-# For details: https://github.com/PyCQA/pylint/blob/master/COPYING
+# For details: https://github.com/PyCQA/pylint/blob/main/LICENSE
 
 import os
 import sys
 import warnings
 
-from pylint import __pkginfo__, config, extensions, interfaces
-from pylint.constants import full_version
+from pylint import __pkginfo__, extensions, interfaces
+from pylint.constants import DEFAULT_PYLINT_HOME, OLD_DEFAULT_PYLINT_HOME, full_version
 from pylint.lint.pylinter import PyLinter
 from pylint.lint.utils import ArgumentPreprocessingError, preprocess_options
-from pylint.utils import utils
+from pylint.utils import print_full_documentation, utils
 
 try:
     import multiprocessing
+    from multiprocessing import synchronize  # noqa pylint: disable=unused-import
 except ImportError:
-    multiprocessing = None  # type: ignore
+    multiprocessing = None  # type: ignore[assignment]
 
 
 def _cpu_count() -> int:
     """Use sched_affinity if available for virtualized or containerized environments."""
     sched_getaffinity = getattr(os, "sched_getaffinity", None)
-    # pylint: disable=not-callable,using-constant-test
+    # pylint: disable=not-callable,using-constant-test,useless-suppression
     if sched_getaffinity:
         return len(sched_getaffinity(0))
     if multiprocessing:
@@ -34,13 +35,13 @@ def cb_list_extensions(option, optname, value, parser):
     for filename in os.listdir(os.path.dirname(extensions.__file__)):
         if filename.endswith(".py") and not filename.startswith("_"):
             extension_name, _, _ = filename.partition(".")
-            print("pylint.extensions.{}".format(extension_name))
+            print(f"pylint.extensions.{extension_name}")
     sys.exit(0)
 
 
 def cb_list_confidence_levels(option, optname, value, parser):
     for level in interfaces.CONFIDENCE_LEVELS:
-        print("%-18s: %s" % level)
+        print(f"%-18s: {level}")
     sys.exit(0)
 
 
@@ -72,9 +73,14 @@ group are mutually exclusive.",
         return 1
 
     def __init__(
-        self, args, reporter=None, exit=True, do_exit=UNUSED_PARAM_SENTINEL,
+        self,
+        args,
+        reporter=None,
+        exit=True,
+        do_exit=UNUSED_PARAM_SENTINEL,
     ):  # pylint: disable=redefined-builtin
         self._rcfile = None
+        self._output = None
         self._version_asked = False
         self._plugins = []
         self.verbose = None
@@ -87,7 +93,9 @@ group are mutually exclusive.",
                     "init-hook": (cb_init_hook, True),
                     "rcfile": (self.cb_set_rcfile, True),
                     "load-plugins": (self.cb_add_plugins, True),
+                    "enable-all-extensions": (self.cb_enable_all_extensions, False),
                     "verbose": (self.cb_verbose_mode, False),
+                    "output": (self.cb_set_output, True),
                 },
             )
         except ArgumentPreprocessingError as ex:
@@ -107,6 +115,17 @@ group are mutually exclusive.",
                         "help": "Specify a configuration file to load.",
                     },
                 ),
+                (
+                    "output",
+                    {
+                        "action": "callback",
+                        "callback": Run._return_one,
+                        "group": "Commands",
+                        "type": "string",
+                        "metavar": "<file>",
+                        "help": "Specify an output file.",
+                    },
+                ),
                 (
                     "init-hook",
                     {
@@ -139,7 +158,8 @@ group are mutually exclusive.",
                         "callback": self.cb_list_messages,
                         "group": "Commands",
                         "level": 1,
-                        "help": "Generate pylint's messages.",
+                        "help": "Display a list of all pylint's messages divided by whether "
+                        "they are emittable with the given interpreter.",
                     },
                 ),
                 (
@@ -150,8 +170,8 @@ group are mutually exclusive.",
                         "callback": self.cb_list_messages_enabled,
                         "group": "Commands",
                         "level": 1,
-                        "help": "Display a list of what messages are enabled "
-                        "and disabled with the given configuration.",
+                        "help": "Display a list of what messages are enabled, "
+                        "disabled and non-emittable with the given configuration.",
                     },
                 ),
                 (
@@ -229,16 +249,6 @@ group are mutually exclusive.",
                         "displayed, and no reports are done by default.",
                     },
                 ),
-                (
-                    "py3k",
-                    {
-                        "action": "callback",
-                        "callback": self.cb_python3_porting_mode,
-                        "help": "In Python 3 porting mode, all checkers will be "
-                        "disabled and only messages emitted by the porting "
-                        "checker will be displayed.",
-                    },
-                ),
                 (
                     "verbose",
                     {
@@ -249,6 +259,15 @@ group are mutually exclusive.",
                         "will be displayed.",
                     },
                 ),
+                (
+                    "enable-all-extensions",
+                    {
+                        "action": "callback",
+                        "callback": self.cb_enable_all_extensions,
+                        "help": "Load and enable all available extensions. "
+                        "Use --list-extensions to see a list all available extensions.",
+                    },
+                ),
             ),
             option_groups=self.option_groups,
             pylintrc=self._rcfile,
@@ -261,7 +280,20 @@ group are mutually exclusive.",
         # load command line plugins
         linter.load_plugin_modules(self._plugins)
         # add some help section
-        linter.add_help_section("Environment variables", config.ENV_HELP, level=1)
+        linter.add_help_section(
+            "Environment variables",
+            f"""
+The following environment variables are used:
+    * PYLINTHOME
+    Path to the directory where persistent data for the run will be stored. If
+not found, it defaults to '{DEFAULT_PYLINT_HOME}' or '{OLD_DEFAULT_PYLINT_HOME}'
+(in the current working directory).
+    * PYLINTRC
+    Path to the configuration file. See the documentation for the method used
+to search for configuration file.
+""",
+            level=1,
+        )
         linter.add_help_section(
             "Output",
             "Using the default text output, the message format is :                          \n"
@@ -295,7 +327,12 @@ group are mutually exclusive.",
         # read configuration
         linter.disable("I")
         linter.enable("c-extension-no-member")
-        linter.read_config_file(verbose=self.verbose)
+        try:
+            linter.read_config_file(verbose=self.verbose)
+        except OSError as ex:
+            print(ex, file=sys.stderr)
+            sys.exit(32)
+
         config_parser = linter.cfgfile_parser
         # run init hook, if present, before loading plugins
         if config_parser.has_option("MASTER", "init-hook"):
@@ -327,15 +364,14 @@ group are mutually exclusive.",
 
         if linter.config.jobs < 0:
             print(
-                "Jobs number (%d) should be greater than or equal to 0"
-                % linter.config.jobs,
+                f"Jobs number ({linter.config.jobs}) should be greater than or equal to 0",
                 file=sys.stderr,
             )
             sys.exit(32)
         if linter.config.jobs > 1 or linter.config.jobs == 0:
             if multiprocessing is None:
                 print(
-                    "Multiprocessing library is missing, " "fallback to single process",
+                    "Multiprocessing library is missing, fallback to single process",
                     file=sys.stderr,
                 )
                 linter.set_option("jobs", 1)
@@ -346,8 +382,21 @@ group are mutually exclusive.",
         # load plugin specific configuration.
         linter.load_plugin_configuration()
 
-        linter.check(args)
-        score_value = linter.generate_reports()
+        # Now that plugins are loaded, get list of all fail_on messages, and enable them
+        linter.enable_fail_on_messages()
+
+        if self._output:
+            try:
+                with open(self._output, "w", encoding="utf-8") as output:
+                    linter.reporter.out = output
+                    linter.check(args)
+                    score_value = linter.generate_reports()
+            except OSError as ex:
+                print(ex, file=sys.stderr)
+                sys.exit(32)
+        else:
+            linter.check(args)
+            score_value = linter.generate_reports()
 
         if do_exit is not UNUSED_PARAM_SENTINEL:
             warnings.warn(
@@ -359,9 +408,13 @@ group are mutually exclusive.",
         if exit:
             if linter.config.exit_zero:
                 sys.exit(0)
+            elif linter.any_fail_on_issues():
+                # We need to make sure we return a failing exit code in this case.
+                # So we use self.linter.msg_status if that is non-zero, otherwise we just return 1.
+                sys.exit(self.linter.msg_status or 1)
+            elif score_value is not None and score_value >= linter.config.fail_under:
+                sys.exit(0)
             else:
-                if score_value and score_value > linter.config.fail_under:
-                    sys.exit(0)
                 sys.exit(self.linter.msg_status)
 
     def version_asked(self, _, __):
@@ -372,6 +425,10 @@ group are mutually exclusive.",
         """callback for option preprocessing (i.e. before option parsing)"""
         self._rcfile = value
 
+    def cb_set_output(self, name, value):
+        """callback for option preprocessing (i.e. before option parsing)"""
+        self._output = value
+
     def cb_add_plugins(self, name, value):
         """callback for option preprocessing (i.e. before option parsing)"""
         self._plugins.extend(utils._splitstrip(value))
@@ -403,7 +460,7 @@ group are mutually exclusive.",
 
     def cb_full_documentation(self, option, optname, value, parser):
         """optik callback for printing full documentation"""
-        self.linter.print_full_documentation()
+        print_full_documentation(self.linter)
         sys.exit(0)
 
     def cb_list_messages(self, option, optname, value, parser):
@@ -426,9 +483,19 @@ group are mutually exclusive.",
             print(check)
         sys.exit(0)
 
-    def cb_python3_porting_mode(self, *args, **kwargs):
-        """Activate only the python3 porting checker."""
-        self.linter.python3_porting_mode()
-
     def cb_verbose_mode(self, *args, **kwargs):
         self.verbose = True
+
+    def cb_enable_all_extensions(self, option_name: str, value: None) -> None:
+        """Callback to load and enable all available extensions"""
+        for filename in os.listdir(os.path.dirname(extensions.__file__)):
+            # pylint: disable=fixme
+            # TODO: Remove the check for deprecated check_docs after the extension has been removed
+            if (
+                filename.endswith(".py")
+                and not filename.startswith("_")
+                and not filename.startswith("check_docs")
+            ):
+                extension_name = f"pylint.extensions.{filename[:-3]}"
+                if extension_name not in self._plugins:
+                    self._plugins.append(extension_name)
